#include "php.h"
#include "zend_compile.h"
#include "php_ini.h"
#include "php_globals.h"
#include "php_orbit.h"
#include "phorb.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

extern void link_shutdown(void);

static zend_class_entry	*corba_class;
static int le_obj = 0;

void phorb_ensure_orb(CORBA_Environment *ev)
{
	int argc = 1;
	char *argv[] = { "php_orbit", NULL };

	if (!OrbSG(orb)) {
		OrbSG(orb) = CORBA_ORB_init(&argc, argv, 
		                            "orbit-local-orb", ev);
		OrbSG(active) = 1;
		OrbSG(counter) = 0;
	}
}

void phorb_shutdown_orb(void)
{
	CORBA_Environment ev;

	if (OrbSG(orb)) {
		CORBA_exception_init(&ev);
		CORBA_ORB_destroy(OrbSG(orb), &ev);
		CORBA_Object_release((CORBA_Object)OrbSG(orb), &ev);
		CORBA_exception_free(&ev);
		OrbSG(orb) = NULL;
		
		/*
		 * CORBA_ORB_destroy doesn't shutdown linc2
		 * and we leak file descriptors
		 */
		link_shutdown();
		ORBit_genuid_fini();
	}
}

static void phorb_free_object(void *val)
{
	CORBA_Object_release((CORBA_Object)val, NULL);
	if (!--OrbSG(counter) && !OrbSG(active))
		phorb_shutdown_orb();
}

gboolean phorb_is_object(zval *val)
{
	return Z_TYPE_P(val) == IS_OBJECT && Z_OBJCE_P(val) == corba_class;
}

zval *phorb_init_object(zval *res)
{
	object_init_ex(res, corba_class);

	return res;
}

ulong phorb_register_object(CORBA_Object obj)
{
	OrbSG(counter)++;

	return zend_list_insert((void *)obj, le_obj);
}

void phorb_init_corba_class(zend_class_entry *objclass)
{
	corba_class = objclass;
	le_obj = phorb_register_destructor(phorb_free_object, NULL);
}

void phorb_fini_corba_class(void)
{
}

static gboolean phorb_invoke(CORBA_Object cobj, ORBit_IMethod *method,
                             zval **z_argv, zval *res)
{
	CORBA_Environment ev;
	CORBA_TypeCode ret_tc;
	gpointer ret = NULL, retptr, *argv, *argv_ptr;
	gboolean retcode = FALSE;
	int i;

	phorb_reset_exception_counter();

	ret_tc = method->ret;
	while(ret_tc && ret_tc->kind == CORBA_tk_alias)
		ret_tc = ret_tc->subtypes[0];

	if (ret_tc != CORBA_OBJECT_NIL && ret_tc->kind != CORBA_tk_void) {
		switch(ret_tc->kind) {
		case CORBA_tk_any:
		case CORBA_tk_sequence:
		case CORBA_tk_array:
			retptr = NULL;
			ret = &retptr;
			break;
		case CORBA_tk_struct:
		case CORBA_tk_union:
			if (!(method->flags & ORBit_I_COMMON_FIXED_SIZE)) {
				retptr = NULL;
				ret = &retptr;
				break;
			}
		default:
			ret = ORBit_small_alloc(method->ret);
		}
	}

	argv = g_new0(gpointer, method->arguments._length);
	argv_ptr = g_new0(gpointer, method->arguments._length);

	for(i = 0; i < method->arguments._length; i++) {
		CORBA_TypeCode tc = method->arguments._buffer[i].tc;
		gint flags = method->arguments._buffer[i].flags;

		if (flags & (ORBit_I_ARG_IN | ORBit_I_ARG_INOUT))
			argv[i] = ORBit_small_alloc(tc);
		else {
			argv[i] = &argv_ptr[i];
			if ((tc->kind == CORBA_tk_any
			     || tc->kind == CORBA_tk_sequence
			     || tc->kind == CORBA_tk_struct
			     || tc->kind == CORBA_tk_union
			     || tc->kind == CORBA_tk_array)
			    && (!(flags & ORBit_I_COMMON_FIXED_SIZE)))
				argv_ptr[i] = NULL;
			else
				argv_ptr[i] = ORBit_small_alloc(tc);
		}
	}

	for(i = 0; i < method->arguments._length; i++) {
		gconstpointer buf = argv[i];
		if (!(method->arguments._buffer[i].flags
		      & (ORBit_I_ARG_IN | ORBit_I_ARG_INOUT)))
			continue;

		if (!phorb_put_zval(&buf,
		                    method->arguments._buffer[i].tc,
		                    z_argv[i])) {
			zend_error(E_ERROR,
			           "Could not marshal parameter `%s' (%d)",
			           method->arguments._buffer[i].name
			           	? method->arguments._buffer[i].name
			           	: "<unknown>", i);
			goto cleanup;
		}
	}

	CORBA_exception_init(&ev);
	ORBit_small_invoke_stub(cobj, method, ret, argv,
	                        CORBA_OBJECT_NIL, &ev);
	if (ev._major == CORBA_SYSTEM_EXCEPTION) {
		CORBA_SystemException *ex = ev._any._value;
		zval *tmp;

		tmp = phorb_system_except(ev._id, ex->minor, ex->completed);

		phorb_throw(tmp);
		goto cleanup;
	} else if (ev._major == CORBA_USER_EXCEPTION) {
		CORBA_TypeCode tc = ev._any._type;
		gconstpointer data = ev._any._value;
		zval *tmp = NULL;

		if (tc)
			tmp = phorb_get_zval(&data, tc);
		if (tmp) {
			convert_to_array(tmp);
			tmp = phorb_user_except(ev._id, tmp);
		} else {
			zend_error(E_WARNING,
			           "Unknown exception of type `%s' received",
			           ev._id);
			tmp = phorb_system_except(
			      	"IDL:omg.org/CORBA/UNKNOWN:1.0", 0,
			      	CORBA_COMPLETED_MAYBE);
		}

		phorb_throw(tmp);
		goto cleanup;
	}
	CORBA_exception_free(&ev);

	if (ret_tc != CORBA_OBJECT_NIL && ret_tc->kind != CORBA_tk_void) {
		gconstpointer retptr;
		zval *tmp;

		switch(ret_tc->kind) {
		case CORBA_tk_any:
		case CORBA_tk_sequence:
		case CORBA_tk_array:
			retptr = *(gpointer *)ret;
			break;
		case CORBA_tk_struct:
		case CORBA_tk_union:
			if (!(method->flags & ORBit_I_COMMON_FIXED_SIZE)) {
				retptr = *(gpointer *)ret;
				break;
			}
		default:
			retptr = ret;
		}

		tmp = phorb_get_zval(&retptr, ret_tc);
		if (!tmp) {
			zend_error(E_ERROR,
			           "Could not demarshal return value");
			goto cleanup;
		}

		zval_dtor(res);
		*res = *tmp;
		FREE_ZVAL(tmp);
	}

	for(i = 0; i < method->arguments._length; i++) {
		gconstpointer retptr;
		gint flags = method->arguments._buffer[i].flags;
		zval *tmp;

		if (!(flags & (ORBit_I_ARG_OUT | ORBit_I_ARG_INOUT)))
			continue;

		if (flags & ORBit_I_ARG_OUT)
			retptr = argv_ptr[i];
		else
			retptr = argv[i];

		tmp = phorb_get_zval(&retptr,
		                     method->arguments._buffer[i].tc);
		if (!tmp) {
			zend_error(E_ERROR,
			           "Could not demarshal parameter `%s' (%d)",
			           method->arguments._buffer[i].name
			           	? method->arguments._buffer[i].name
			           	: "<unknown>", i);
			goto cleanup;
		}

		REPLACE_ZVAL_VALUE(&z_argv[i], tmp, 0);
		FREE_ZVAL(tmp);
	}

	retcode = TRUE;

cleanup:
	if (ret) {
		switch(ret_tc->kind) {
		case CORBA_tk_any:
		case CORBA_tk_sequence:
		case CORBA_tk_array:
			CORBA_free(retptr);
			break;
		case CORBA_tk_struct:
		case CORBA_tk_union:
			if (!(method->flags & ORBit_I_COMMON_FIXED_SIZE)) {
				CORBA_free(retptr);
				break;
			}
		default:
			CORBA_free(ret);
		}
	}

	if (argv) {
		for(i = 0; i < method->arguments._length; i++) {
			if (method->arguments._buffer[i].flags
			    & ORBit_I_ARG_OUT)
				CORBA_free(argv_ptr[i]);
			else
				CORBA_free(argv[i]);
		}

		g_free(argv);
		g_free(argv_ptr);
	}

	if (phorb_reset_exception_counter()) {
		phorb_catch();
		return FALSE;
	}

	return retcode;
}

CORBA_char *phorb_get_object_repoid(CORBA_Object cobj, CORBA_Environment *ev)
{
	struct CORBA_Object_internal {
		struct ORBit_RootObject_struct	root;
		void				*reserved;
		GQuark				type_qid;
	};
	CORBA_Environment ev_;
	const char *repoid = NULL;

	if (!ev) {
		ev = &ev_;
		CORBA_exception_init(ev);
	}

	repoid = g_quark_to_string(
	         	((struct CORBA_Object_internal *)cobj)->type_qid);
	if (ev == &ev_)
		CORBA_exception_free(ev);

	if (!repoid)
		return NULL;

	return CORBA_string_dup(repoid);
}

void phorb_invoke_method(CORBA_Object cobj, char *method, zval **argv,
                         int argc, zval *res)
{
	CORBA_Environment ev;
	phorbIfaceInfo *info = NULL;
	ORBit_IMethod *imethod;
	CORBA_char *repo_id;
	int i;
	TSRMLS_FETCH();

	CORBA_exception_init(&ev);

	repo_id = phorb_get_object_repoid(cobj, &ev);
	if (repo_id) {
		info = phorb_find_interface_description(NULL, repo_id);
		CORBA_free(repo_id);
	}

	if (!info) {
		zend_error(E_WARNING, "Invocation on undefined interface");
		goto leave;
	}

	imethod = phorb_find_interface_method(info, method);
	if (!imethod) {
		zend_error(E_WARNING,
		           "Invocation on undefined method `%s' in "
		           "interface `%s'", method, info->iface->tc->repo_id);
		goto leave;
	}

	if (argc != imethod->arguments._length) {
		WRONG_PARAM_COUNT;
		goto leave;
	}
	for(i = 0; i < argc; i++) {
		if ((imethod->arguments._buffer[i].flags
		     & (ORBit_I_ARG_OUT | ORBit_I_ARG_INOUT))
		    && !PZVAL_IS_REF(argv[i])) {
			zend_error(E_WARNING, "Parameter %d must be passed "
			                      "by reference", i);
			goto leave;
		}
	}

	if (!phorb_invoke(cobj, imethod, argv, res) < 0)
		goto leave;

  leave:
	CORBA_exception_free(&ev);
	return;
}

void phorb_get_attribute(CORBA_Object cobj, const char *name, zval *res)
{
	ORBit_IMethod *imethod;
	CORBA_char *repo_id = NULL;
	phorbIfaceInfo *info = NULL;
	char *method;
	int type;

	repo_id = phorb_get_object_repoid(cobj, NULL);
	if (repo_id)
		info = phorb_find_interface_description(NULL, repo_id);

	if (strcmp(name, "_tc") == 0) {
		CORBA_TypeCode tc;
		zval *val;

		tc = phorb_find_typecode(NULL, repo_id);
		if (!tc)
			tc = TC_CORBA_Object;

		res = phorb_init_typecode(res);
		if (!res)
			goto out;

		MAKE_STD_ZVAL(val);
		ZVAL_RESOURCE(val,
		              phorb_register_typecode(
		              	(CORBA_TypeCode)CORBA_Object_duplicate(
		              		(CORBA_Object)tc, NULL)));
		zend_hash_index_update(Z_OBJPROP_P(res), 0, &val,
		                       sizeof(zval *), NULL);

		goto out;
	} else if (strcmp(name, "_hash") == 0) {
		CORBA_Environment ev;

		CORBA_exception_init(&ev);
		ZVAL_LONG(res, CORBA_Object_hash(cobj, LONG_MAX, &ev));
		CORBA_exception_free(&ev);

		goto out;
	}

	repo_id = phorb_get_object_repoid(cobj, NULL);
	if (repo_id)
		info = phorb_find_interface_description(NULL, repo_id);

	if (!info) {
		zend_error(E_WARNING, "Get invocation on undefined interface");
		goto out;
	}

	method = g_strdup_printf("_get_%s", name);
	imethod = phorb_find_interface_method(info, method);
	g_free(method);
	if (!imethod || imethod->arguments._length != 0
	    || imethod->ret == CORBA_OBJECT_NIL
	    || imethod->ret->kind == CORBA_tk_void) {
		zend_error(E_WARNING, "Get invocation on undefined name "
		           "%s in interface %s", name,
		           info->iface->tc->repo_id);
		goto out;
	}

	if (!phorb_invoke(cobj, imethod, NULL, res))
		goto out;

out:
	if (repo_id)
		CORBA_free(repo_id);
}

int phorb_set_attribute(CORBA_Object cobj, const char *name, zval *value)
{
	zval *argv[1] = { value };
	ORBit_IMethod *imethod;
	CORBA_char *repo_id;
	gchar *method;
	phorbIfaceInfo *info = NULL;
	int type;

	repo_id = phorb_get_object_repoid(cobj, NULL);
	if (repo_id) {
		info = phorb_find_interface_description(NULL, repo_id);
		CORBA_free(repo_id);
	}

	if (!info) {
		zend_error(E_WARNING, "Set invocation on undefined interface");
		return FAILURE;
	}

	method = g_strdup_printf("_set_%s", name);
	imethod = phorb_find_interface_method(info, method);
	g_free(method);

	if (!imethod || imethod->arguments._length != 1
	    || (imethod->ret != CORBA_OBJECT_NIL
	        && imethod->ret->kind != CORBA_tk_void)) {
		zend_error(E_WARNING, "Set invocation on undefined attribute "
		           "%s in interface %s", name,
		           info->iface->tc->repo_id);
		return FAILURE;
	}

	SEPARATE_ZVAL(&argv[0]);
	if (!phorb_invoke(cobj, imethod, argv, NULL))
		return FAILURE;

	return SUCCESS;
}
